/*
 * Copyright (c) 2008 Todd C. Miller <Todd.Miller@courtesan.com>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <config.h>

#include <sys/types.h>
#include <err.h>
#include <errno.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "compat.h"

/*
 * Return codes:
 * EXIT_FAILURE ... unspecified error
 *    0 ... everything ok
 *   30 ... invalid -e arg value
 *   31 ... odd number of paths
 *   32 ... copy operation failed, no files copied
 *   33 ... copy operation failed, some files copied
 */

int
main (int argc, char **argv)
{
     char *cp, *cmnd;

     if (argc < 2)
	  errx(EXIT_FAILURE, "requires at least one argument");

     /* check the first argument, if it's `-e' then we are in sudoedit mode */
     if (strncmp(argv[1], "-e", 3) == 0) {
	  int fd_src, fd_dst, post, n, ret = -1;
	  ssize_t nread, nwritten;
	  char *path_src, *path_dst, buf[BUFSIZ];

	  if (argc < 3)
	       return EXIT_FAILURE;

	  /*
	   * We need to know whether we are performing the copy operation
	   * before or after the editing. Without this we would not know
	   * which files are temporary and which are the originals.
	   *  post = 0 ... before
	   *  post = 1 ... after
	   */
	  if (strncmp(argv[2], "0", 2) == 0)
	       post = 0;
	  else if (strncmp(argv[2], "1", 2) == 0)
	       post = 1;
	  else /* invalid value */
	       return 30;

	  /* align argv & argc to the beggining of the file list */
	  argv += 3;
	  argc -= 3;

	  /* no files specified, nothing to do */
	  if (argc == 0)
	       return 0;
	  /* odd number of paths specified */
	  if (argc % 2 == 1)
	       return 31;

	  for (n = 0; n < argc - 1; n += 2) {
	       path_src = argv[n];
	       path_dst = argv[n+1];
	       /*
		* Try to open the source file for reading. If it
		* doesn't exist, it's ok, we'll create an empty
		* destination file.
		*/
	       if ((fd_src = open(path_src, O_RDONLY, 0600)) < 0) {
		    if (errno == ENOENT) {
			 /* new file */
		    } else {
			 warn("open(%s)", path_src);
			 if (post) {
			      ret = 33;
			      goto nocleanup;
			 } else
			      goto cleanup_0;
		    }
	       }

	       /*
		* Use O_EXCL if we are not in the post editing stage
		* so that it's ensured that the temporary files are
		* created by us and that we are not opening any sym-
		* links.
		*/
	       if ((fd_dst = open(path_dst, (post ? 0 : O_EXCL) |
				  O_WRONLY|O_TRUNC|O_CREAT, post ? 0644 : 0600)) < 0)
	       {
		    /* error - cleanup */
		    warn("open(%s%s)", path_dst, post ? "" : ", O_EXCL");
		    if (post) {
			 ret = 33;
			 goto nocleanup;
		    } else
			 goto cleanup_0;
	       }

	       if (fd_src != -1) {
		    while ((nread = read(fd_src, buf, sizeof(buf))) > 0) {
			 if ((nwritten = write(fd_dst, buf, nread)) != nread) {
			      warn("write");
			      if (post) {
				   ret = 33;
				   goto nocleanup;
			      } else
				   goto cleanup_0;
			 }
		    }
	       }

	       if (fd_dst != -1)
		    close(fd_dst);
	       if (fd_src != -1)
		    close(fd_src);
	       fd_dst = fd_src = -1;
	  }

	  ret = 0;
	  /* remove temporary files (post=1) */
	  for (n = 0; n < argc - 1; n += 2)
	       unlink(argv[n]);	  
     nocleanup:
	  if (fd_dst != -1)
	       close(fd_dst);
	  if (fd_src != -1)
	       close(fd_src);
	  _exit(ret);
     cleanup_0:
	  /* remove temporary files (post=0) */
	  for (n = 0; n < argc - 1; n += 2)
	       unlink(argv[n+1]);
	  if (fd_dst != -1)
	       close(fd_dst);
	  if (fd_src != -1)
	       close(fd_src);
	  _exit(32);
     } else {
	  /* Shift argv and make a copy of the command to execute. */
	  argv++;
	  argc--;
	  cmnd = strdup(argv[0]);
	  if (cmnd == NULL)
	       err(EXIT_FAILURE, NULL);

	  /* If invoked as a login shell, modify argv[0] accordingly. */
	  if (argv[0][0] == '-') {
	       if ((cp = strrchr(argv[0], '/')) == NULL)
		    cp = argv[0];
	       *cp = '-';
	  }
	  execv(cmnd, argv);
	  warn("unable to execute %s", argv[0]);
     }
     _exit(EXIT_FAILURE);
}
